<?php

namespace Drupal\commerce_gc_client\Controller;

use Drupal\Component\Serialization\Json;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\RedirectResponse;
use GuzzleHttp\Exception\RequestException;

/**
 * Functions for communicating with the GoCardless partner: Seamless-CMS.co.uk.
 */
class GoCardlessPartner {

  protected $session;

  /**
   * Constructs the object.
   *
   * @param string $gateway_id
   *   The GoCardless payment gateway entity ID.
   */
  public function __construct($gateway_id = NULL) {
    $this->settings = $this->getSettings($gateway_id);
    $this->session = \Drupal::request()->getSession();
  }

  /**
   * Combines module settings and gateway settings into a single array.
   */
  public static function getSettings($gateway_id) {

    if ($gateway_id) {
      $module_settings = \Drupal::config('commerce_gc_client.settings')->get();
      $gateway_settings = \Drupal::config('commerce_payment.commerce_payment_gateway.' . $gateway_id)->get();
      return $gateway_settings + $module_settings;
    }
    else {
      return \Drupal::config('commerce_gc_client.settings')->get();
    }
  }

  /**
   * Handles API posts to GC Partner site, and optionally logs results.
   */
  public function api($params) {

    if (isset($params['mode'])) {
      $this->mode = $params['mode'];
      $auth = $this->authenticate();
      if (count($params) == 1) {
        return $auth;
      }
    }
    elseif (isset($this->settings['configuration'])) {
      $this->mode = $this->settings['configuration']['mode'];
    }

    $auth = $this->authenticate();
    if ($auth == 200) {
      $result = $this->post($params);
      if (isset($result->error)) {
        $message = t('Error calling GoCardless API, code @code (@error)', [
          '@code' => $result->code,
          '@error' => $result->error,
        ]);
        drupal_set_message($message, 'error');
        if ($this->settings['configuration']['log_api']) {
          \Drupal::logger("commerce_gc_client")->error('<pre>' . $message . '<br />' . print_r($result, TRUE) . '</pre>', []);
        }
        $result->response = new stdClass();
        $result->response->status_code = $result->code;
        return $result;
      }
      else {
        if (isset($this->settings['configuration']) && $this->settings['configuration']['log_api']) {
          \Drupal::logger("commerce_gc_client")->notice('<pre>GoCardless API response: <br />' . print_r($result, TRUE) . '</pre>', []);
        }
        return $result;
      }
    }
    else {
      drupal_set_message(t('Error @code connecting with partner site', [
        '@code' => $auth,
      ]), 'error');
      if (\Drupal::config('commerce_gc_client.settings')->get('commerce_gc_client_debug_api')) {
        \Drupal::logger("commerce_gc_client")->error('<pre>' . print_r($auth, TRUE) . '</pre>', []);
      }
      return $auth;
    }
  }

  /**
   * Handles session authentication with GC Partner site.
   */
  protected function authenticate() {

    if ($this->session->get('gc_client_cookie_created') && $this->session->get('gc_client_cookie_created') < REQUEST_TIME - 1800) {
      $this->session->remove('gc_client_cookie');
      $this->session->remove('gc_client_cookie_created');
    }

    if (!$this->session->get('gc_client_cookie_created')) {
      $user_name = $this->settings['partner_user_' . $this->mode];
      $user_pass = $this->settings['partner_pass_' . $this->mode];

      // Attempt session authentication if user name and password set.
      if (isset($user_name) && isset($user_pass)) {

        // Login first.
        $data = [
          'username' => $user_name,
          'password' => $user_pass,
        ];
        $data = Json::encode($data);
        $url = $this->settings['partner_url'] . '/gc_connect/user/login';

        try {
          $result = \Drupal::httpClient()
            ->post($url, [
              'headers' => ['Content-Type' => 'application/json'],
              'body' => $data,
            ]
          );
          $result_data = (string) $result->getBody();
          if (empty($result_data)) {
            return FALSE;
          }
          $result_data = json_decode($result_data);
        }
        catch (RequestException $e) {
          return FALSE;
        }

        if ($result->getStatusCode() == 200) {

          // Get X-CRSF token, and save cookie and token.
          $this->session->set('gc_client_cookie', $result_data->session_name . '=' . $result_data->sessid);
          $this->session->set('gc_client_cookie_created', REQUEST_TIME);

          $xcrf_url = $this->settings['partner_url'] . '/services/session/token';

          try {
            $xcrf_result = \Drupal::httpClient()
              ->get($xcrf_url, [
                'headers' => ['Cookie' => $this->session->get('gc_client_cookie')],
              ]
            );
            $xcrf_result_data = (string) $xcrf_result->getBody();
            if (empty($xcrf_result_data)) {
              return FALSE;
            }
            $this->session->set('gc_client_token', $xcrf_result_data);
          }
          catch (RequestException $e) {
            return FALSE;
          }
        }
        return $result->getStatusCode();
      }
      else {
        return $result = 'User name and password not set for ' . $this->mode;
      }
    }
    else {
      // Already logged in.
      return $result = 200;
    }
  }

  /**
   * Handles HTTP requests to GC Partner site.
   */
  private function post($params) {

    $headers = [
      'Content-Type' => 'application/json',
      'Cookie' => $this->session->get('gc_client_cookie'),
      'X-CSRF-Token' => $this->session->get('gc_client_token'),
    ];

    $params['environment'] = strtoupper($this->mode);
    $params = json_encode($params);
    $url = $this->settings['partner_url'] . '/gc/client/' . $this->settings['org_id_' . $this->mode];

    try {
      $response = \Drupal::httpClient()
        ->post($url, [
          'headers' => $headers,
          'body' => $params,
        ]
      );
      $body = (string) $response->getBody();
      if (empty($body)) {
        return FALSE;
      }
      return $body = json_decode($body);
    }
    catch (RequestException $e) {
      return FALSE;
    }

    if (isset($response->error)) {
      return $response;
    }
  }

  /**
   * Callback function: Saves key variables for connecting with Partner site.
   *
   * Variables are posted here from Partner site, following completion of
   * GoCardless OAuth flow.
   */
  public static function connect() {

    if (isset($_POST['environ'])) {

      $_POST['environ'] == 'SANDBOX' ? $ext = '_sandbox' : $ext = '_live';
      $settings = \Drupal::service('config.factory')->getEditable('commerce_gc_client.settings');

      // Check that the  returned password matches the saved one before
      // proceeding, and return error 403 Forbidden if it does not.
      if (isset($_POST['pass']) && $_POST['pass'] == $settings->get('partner_pass' . $ext)) {
        if (isset($_POST['id'])) {
          $settings->set('org_id' . $ext, $_POST['id'])->save();
        }
        if (isset($_POST['name'])) {
          $settings->set('partner_user' . $ext, $_POST['name'])->save();
        }
      }
      else {
        return new Response('Forbidden', 403, [
          'Content-Type' => 'text/html',
        ]);
      }
    }

    // Unset to remove possibility of authentication failure if client is
    // reconnecting.
    if ($this->session->get('gc_client_cookie_created')) {
      $this->session->remove('gc_client_cookie_created');
    }

    return new Response();
  }

  /**
   * Redirects user to GC settings page upon completion of OAuth flow.
   */
  public static function connectComplete() {

    if (isset($_GET['status'])) {

      if ($_GET['status'] == 'insecure') {
        drupal_set_message(t('Connection cannot be created because site must be secure (https) to use LIVE environment'), 'error');
      }
      elseif ($_GET['status'] == 'connected') {
        drupal_set_message(t('You have connected successfully with GoCardless'));
      }
      // Status will be failed if the wrong passward is provided in the request.
      elseif ($_GET['status'] == 'failed') {
        drupal_set_message(t('Connection with GoCardless has failed'), 'error');
      }
    }

    if (isset($_GET['gateway_id'])) {
      $gateway_id = $_GET['gateway_id'];
      return new RedirectResponse('/admin/commerce/config/payment-gateways/manage/' . $gateway_id);
    }
  }

}
