<?php

namespace Drupal\commerce_gc_client\Form;

use Drupal\Core\Locale\CountryManager;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\FormBase;
use Drupal\commerce_price\Entity\Currency;

/**
 * Defines a form for configurring GoCardless currencies.
 */
class Currencies extends FormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'commerce_gc_client_currencies_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $currencies = \Drupal::config('commerce_gc_client.settings')->get('currency_schemes');

    // TODO find easier way of loading enabled currencies.
    foreach (\Drupal::configFactory()->listAll('commerce_price.commerce_currency') as $commerce_currency) {
      $commerce_currency_code = substr($commerce_currency, -3);
      $commerce_currencies[$commerce_currency_code] = Currency::load($commerce_currency_code);
    }

    $countries = CountryManager::getStandardList();
    $rows = [];
    $rows_unavailable = [];
    $defaults = [];
    foreach ($currencies as $currency_code => & $currency) {

      $currency_countries = [];
      foreach ($currency['countries'] as $country_code => $country) {

        $country_name = $countries[$country_code];
        if (isset($country['failure_warning'])) {
          $country_name .= '*';
        }
        $currency_countries[$country_code] = $country_name;
      }
      $currency_countries = implode(', ', $currency_countries);

      $row = [
        $currency_code,
        $currency['scheme'],
        $currency_countries,
      ];

      if (isset($commerce_currencies[$currency_code])) {
        $rows[$currency_code] = $row;
        if (isset($currency['enabled']) && $currency['enabled']) {
          $defaults[$currency_code] = TRUE;
        }
      }
      else {
        $rows_unavailable[$currency_code] = $row;
      }
    }
    $header = [t('Currency code'), t('Scheme'), t('Country / countries')];

    $markup = '<p>' . t('For information on currencies and countries that are serviced by GoCardless <a target="new" href ="https://gocardless.com/faq/merchants/international-payments">view their FAQ page here</a>. To collect payments from multiple regions the relevant direct debit schemes need to be enabled by GoCardless. See <a target="new" href="https://support.gocardless.com/hc/en-gb/articles/115002833785-Collecting-from-multiple-regions">GoCardless article</a> for more information on how to do this."') . '</p>';

    $form['markup'] = ['#markup' => $markup];

    $form['currencies'] = [
      '#type' => 'tableselect',
      '#header' => $header,
      '#options' => $rows,
      '#empty' => t('No countries available.'),
      '#default_value' => $defaults,
    ];

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => t('Enable / disable currencies'),
    ];

    $form['markup_unavailable'] = [
      '#markup' => '<p>' . t('GoCardless currencies that are currently unavailable:') . '</p>',
    ];

    $form['currencies_unavailable'] = [
      '#theme' => 'table',
      '#header' => $header,
      '#rows' => $rows_unavailable,
      '#empty' => t('All currencies available.'),
    ];

    $form['markup_failure_warning'] = [
      '#markup' => '<p style = "font-size: smaller;">' . t('* Direct Debits in these countries have a relatively high failure rate.') . '</p>',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $settings = \Drupal::configFactory()->getEditable('commerce_gc_client.settings');
    $currencies = $settings->get('currency_schemes');
    $values = $form_state->getValue(['currencies']);
    foreach ($currencies as $currency_code => $currency) {
      $enabled = FALSE;
      if (isset($values[$currency_code]) && $values[$currency_code]) {
        $values[$currency_code] ? $enabled = TRUE : NULL;
      }
      $currencies[$currency_code]['enabled'] = $enabled;
    }
    $settings->set('currency_schemes', $currencies)->save();
    drupal_set_message(t('Available GoCardless currencies have been updated.'));
  }

}
