<?php

namespace Drupal\commerce_gc_client\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\commerce_order\Entity\OrderItem;
use Drupal\commerce_gc_client\Controller\GoCardlessPartner;
use Drupal\commerce_gc_client\Event\GoCardlessEvents;
use Drupal\commerce_gc_client\Event\PaymentCreatedEvent;
use Drupal\commerce_price\Entity\Currency;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Ajax\RedirectCommand;
use Drupal\Core\Ajax\AjaxResponse;

/**
 * Defines a form for managing GoCardless mandates for order items.
 */
class Mandate extends FormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'commerce_gc_client_mandate_form';
  }

  /**
   * Form constructor.
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $order = \Drupal::routeMatch()->getParameter('commerce_order');
    $this->order_id = $order->id();
    $this->db = \Drupal::database();
    $session = \Drupal::service('user.private_tempstore')->get('commerce_gc_client');

    $query = $this->db->select('commerce_gc_client', 'g');
    $query->join('commerce_gc_client_item', 'i', 'i.gcid=g.gcid');
    $this->gcs = $query
      ->fields('g')
      ->fields('i')
      ->condition('order_id', $this->order_id)
      ->execute()
      ->fetchAllAssoc('item_id');

    $count = count($this->gcs);
    if ($count == 0) {
      drupal_set_message(t('You have no GoCardless mandates associated with this order.'), 'warning');
      return;
    }

    $payment_gateway_id = $order->get('payment_gateway')->first()->entity->Id();
    $this->partner = new GoCardlessPartner($payment_gateway_id);
    $user = \Drupal::currentUser();
    $this->commerce_currency_code = $order->getTotalPrice()->getCurrencyCode();
    $commerce_currency_symbol = Currency::load($this->commerce_currency_code)->getSymbol();

    foreach (\Drupal::config('commerce_gc_client.settings')->get('currency_schemes') as $this->currency_code => $this->gc_currency) {
      if ($this->gc_currency['scheme'] == reset($this->gcs)->gc_mandate_scheme) {
        $currency_symbol = Currency::load($this->currency_code)->getSymbol();
        break;
      }
    }

    // Create an array of unique line item names.
    $rows = [];
    $this->item_select = [];
    $title_count = [];
    foreach ($order->getItems() as $item) {

      $select_id = $item->id();
      $item_title = $item->getTitle();
      !isset($title_count[$item_title]) ? $title_count[$item_title] = 1 : $title_count[$item_title]++;
      $title_count[$item_title] == 1 ? $title[$select_id] = $item_title : $title[$select_id] = $item_title . ' (' . $title_count[$item_title] . ')';
      $this->item_select[$select_id] = $title[$select_id];
    }

    if ($count == 1) {
      $this->item_id = reset($this->gcs)->item_id;
      $text_top = t('<h3><b>Administrate <a href = "/admin/commerce/orders/@order_id">@title</a></b></h3>', [
        '@title' => $item_title,
        '@order_id' => $this->order_id,
      ]);
      $form['text_top'] = [
        '#type' => 'item',
        '#markup' => $text_top,
      ];
    }
    else {
      if ($form_state->getUserInput()) {
        $this->item_id = $form_state->getUserInput()['item_id'];
      }
      elseif ($this->item_id = $session->get('item_id_' . $this->order_id)) {
      }
      else {
        $this->item_id = reset($this->gcs)->item_id;
      }

      $form = ['#attributes' => ['id' => ['switch-item']]];

      $text_top = t('<h3><b>Administrate <a href = "@path">@title</a></b></h3>', [
        '@title' => $this->item_select[$this->item_id],
        '@path' => '/admin/store/orders/' . $this->order_id,
      ]);

      $form['text_top'] = ['#type' => 'item', '#markup' => $text_top];
      $form['item_id'] = ['#type' => 'value', '#value' => $this->item_id];
      $form['order_id'] = ['#type' => 'value' , '#value' => $this->order_id];

      $form['item_id'] = [
        '#title' => t('Switch item'),
        '#type' => 'select',
        '#options' => $this->item_select,
        '#default_value' => $this->item_id,
        '#ajax' => [
          'callback' => [$this, 'switchItem'],
          'wrapper' => 'switch-item',
          'method' => 'replace',
        ],
      ];
    }

    $this->gc = $this->gcs[$this->item_id];
    $result = $this->partner->api([
      'endpoint' => 'mandates',
      'action' => 'get',
      'mandate' => $this->gc->gc_mandate_id,
    ]);
    if ($result->response->status_code != 200) {
      drupal_set_message(t('No active mandate found for this order at GoCardless.'), 'warning');
      return;
    }
    else {
      $this->mandate = $result->response->body->mandates;
    }

    $this->item = OrderItem::load($this->item_id);
    $total = $this->item->getTotalPrice()->getNumber();
    $this->item->getData('gc') ? $data = $this->item->getData('gc') : $data = NULL;
    $this->default_amount = number_format((float) $total, 2, '.', '');

    $interval = isset($data['interval_params']) ? $data['interval_params']['string'] : NULL;

    if (isset($data['shipment'])) {
      foreach ($order->getAdjustments() as $adj) {
        if ($adj->getType() == 'shipping') {
          $adj_amount = $adj->getAmount()->getNumber();
          $shipment_currency_code = $adj->getAmount()->getCurrencyCode();
          $shipment_amount = round($adj_amount * $data['shipment']['proportion'], 2);
          $this->default_amount += $shipment_amount;
          $shipment_arr = [
            'Shipping' => \Drupal::service('commerce_price.currency_formatter')->format($shipment_amount, $shipment_currency_code),
          ];
          break;
        }
      }
    }

    $result = $this->partner->api([
      'endpoint' => 'payments',
      'action' => 'list',
      'mandate' => $this->gc->gc_mandate_id,
      'limit' => 500,
    ]);
    if ($result->response->status_code == 200) {
      $payments = $result->response->body->payments;
    }

    // Data for one-off payments.
    if ($this->gc->type == 'P') {

      $item_data = [
        'Mandate ID' => $this->mandate->id,
        'Mandate status' => ucfirst($this->mandate->status),
        'Created' => format_date(strtotime($this->mandate->created_at), 'short'),
        'Type' => t('Payments'),
        'Amount' => \Drupal::service('commerce_price.currency_formatter')->format($total, $this->commerce_currency_code),
      ];
      if (isset($shipment_arr)) {
        $item_data += $shipment_arr;
      }
      $item_data += [
        'Interval' => $interval == ' ' ? '-' : $interval,
        'Next possible charge date' => !is_null($this->mandate->next_possible_charge_date) ? format_date(strtotime($this->mandate->next_possible_charge_date), 'short') : '-',
        'Next scheduled payment creation' => $this->mandate->status != 'cancelled' && !is_null($this->gc->next_payment) ? format_date($this->gc->next_payment, 'short') : '-',
      ];

      $this->default_name = $this->item_select[$this->item_id];

      // Data for scheduled adjustments table.
      $query = $this->db->select('commerce_gc_client_item', 'i');
      $query->join('commerce_gc_client_item_schedule', 's', 's.item_id=i.item_id');
      $adjustments = $query->fields('s')->condition('i.item_id', $this->gc->item_id)->condition('s.type', 'adj')->orderBy('timestamp', 'ASC')->execute()->fetchAll();

      if (!empty($adjustments)) {

        $adj_headers = [t('Title'), t('Adjustment'), t('Date'), t('Status'), ''];
        $adj_rows = [];
        foreach ($adjustments as $adj) {

          $actions = '';
          if ($adj->status == 1) {
            $actions = t('<a href="@cancel_path">Cancel</a> @spacer', [
              '@cancel_path' => '/gc_client/adjustment_action/cancel/' . $this->order_id . '/' . $adj->sid,
              '@spacer' => '| ',
            ]);
          }
          $actions .= t('<a href="@delete_path">Delete</a>', [
            '@delete_path' => '/gc_client/adjustment_action/delete/' . $this->order_id . '/' . $adj->sid,
          ]);

          if ($adj->status == 0) {
            $status = 'Cancelled';
          }
          elseif ($adj->status == 1) {
            $status = 'Pending';
          }
          elseif ($adj->status == 2) {
            $status = 'Complete';
          }

          $adj_data = unserialize($adj->data);
          $adj_rows[] = [
            $adj_data['title'],
            \Drupal::service('commerce_price.currency_formatter')->format($adj_data['amount'], $this->commerce_currency_code),
            $adj->date,
            $status,
            new FormattableMarkup($actions, []),
          ];
        }
      }
    }

    // Data for subscriptions table.
    elseif ($this->gc->type == 'S') {

      $this->subscription = [];
      if ($this->gc->gc_subscription_id) {
        $result = $this->partner->api([
          'endpoint' => 'subscriptions',
          'action' => 'get',
          'id' => $this->gc->gc_subscription_id,
        ]);
        if ($result->response->status_code == 200) {
          $this->subscription = $result->response->body->subscriptions;
        }
      }

      if (!count($this->subscription) || !$this->gc->gc_subscription_id) {
        drupal_set_message(t('No subscription found at GoCardless for this line item.'), 'warning');
        return $form;
      }

      $item_data = [
        'Mandate ID' => $this->mandate->id,
        'Mandate status' => ucfirst($this->mandate->status),
        'Created' => format_date(strtotime($this->mandate->created_at), 'gocardless_client'),
        'Type' => t('Subscription'),
        'Subscription status' => ucfirst($this->subscription->status),
        'Subscription name' => $this->subscription->name,
        'Subscription ID' => $this->subscription->id,
        'Subscription start' => format_date(strtotime($this->subscription->start_date), 'gocardless_client'),
        'Amount' => \Drupal::service('commerce_price.currency_formatter')->format($total, $this->commerce_currency_code),
      ];
      if (isset($shipment_arr)) {
        $item_data += $shipment_arr;
      }
      $item_data += [
        'Interval' => $interval,
        'Day of month' => $this->subscription->day_of_month,
        'Month' => $this->subscription->month,
        'End date' => !is_null($this->subscription->end_date) ? format_date(strtotime($this->subscription->end_date), 'gocardless_client') : NULL,
      ];

      // Assemble data for upcoming payments table.
      $upcoming_rows = [];
      foreach ($this->subscription->upcoming_payments as $upcoming) {
        $upcoming_rows[] = [
          format_date(strtotime($upcoming->charge_date), 'gocardless_client'),
          \Drupal::service('commerce_price.currency_formatter')->format($upcoming->amount / 100, $this->currency_code),
        ];
      }

      $this->default_name = $this->subscription->name;
    }

    $item_rows = [];
    foreach ($item_data as $item_data_key => $item_data_data) {
      if (is_null($item_data_data)) {
        continue;
      }
      $item_rows[] = [
        t('@key', ['@key' => $item_data_key]),
        $item_data_data,
      ];
    }

    $payment_rows = [];
    $payment_total = 0;
    $payment_statuses = ['confirmed', 'paid_out'];

    if (isset($payments)) {
      foreach ($payments as & $payment) {
        if (isset($payment->links->subscription) && $payment->links->subscription != $this->gc->gc_subscription_id || isset($payment->metadata->item_id) && $payment->metadata->item_id != $this->item_id) {
          continue;
        }

        if ($payment->status == 'pending_submission') {
          $path = '/gc_client/payment_cancel/' . $this->order_id . '/' . $payment->id;
          $link = t('<a href="@path">Cancel</a>', ['@path' => $path]);
        }
        else {
          $link = NULL;
        }

        $payment_rows[$payment->id] = [
          format_date(strtotime($payment->created_at), 'short'),
          $payment->description,
          $payment->id,
          \Drupal::service('commerce_price.currency_formatter')->format($payment->amount / 100, $this->currency_code),
          $payment->status,
          format_date(strtotime($payment->charge_date), 'gocardless_client'),
          $link,
        ];
        if (in_array($payment->status, $payment_statuses)) {
          $payment_total = $payment_total + $payment->amount;
        }
      }
    }

    $item_header = [[
      'data' => 'Item data',
      'colspan' => 2,
      'style' => 'text-align:center;font-weight:bold;',
    ],
    ];

    $form['table'] = [
      '#type' => 'table',
      '#header' => $item_header,
      '#rows' => isset($item_rows) ? $item_rows : NULL,
      '#empty' => t('There are no GoCardless items for this order'),
      '#prefix' => '<div id="gc-item-table">',
      '#suffix' => '</div>',
      '#attached' => ['library' => ['commerce_gc_client/gocardless-client']],
    ];

    if ($this->gc->type == 'S') {
      $form['upcoming_payments'] = [
        '#type' => 'details',
        '#title' => t('Upcoming payments'),
        '#open' => FALSE,
        '#prefix' => '<div id="gc-upcoming-payments-details">',
        '#suffix' => '</div>',
      ];
      $upcoming_header = [t('Charge date'), t('Amount')];
      $form['upcoming_payments']['upcoming_payments_table'] = [
        '#theme' => 'table',
        '#header' => $upcoming_header,
        '#rows' => isset($upcoming_rows) ? $upcoming_rows : NULL,
        '#empty' => t('There are no upcoming payments for this subscription.'),
      ];
    }

    if (isset($payment_rows)) {

      $form['payment_tables'] = [
        '#type' => 'details',
        '#title' => t('Payments'),
        '#description' => t('(Live data from GoCardless.com)'),
        '#open' => TRUE,
      ];
      $payment_header = [
        t('Created'),
        t('Description'),
        t('Payment ID'),
        t('Amount'),
        t('Status'),
        t('Charge customer at'),
        '',
      ];
      $form['payment_tables']['payments_table'] = [
        '#theme' => 'table',
        '#header' => $payment_header,
        '#rows' => isset($payment_rows) ? $payment_rows : NULL,
        '#empty' => t('There are no payments for this item yet.'),
        '#suffix' => t('Total confirmed payments: @payment_total', [
          '@payment_total' => \Drupal::service('commerce_price.currency_formatter')->format($payment_total, $this->currency_code),
        ]),
      ];
    }

    // Create payment section.
    if (in_array($this->mandate->status, [
      'pending_submission',
      'submitted',
      'active',
    ])) {
      $form['create_payment'] = [
        '#type' => 'details',
        '#title' => t('Create a payment'),
        '#open' => FALSE,
      ];
      if ($this->gc->type == 'S') {
        $form['create_payment']['#description'] = t('You can create one-off payments under the existing mandate for this item, which are in addition to the subscription plan listed here.');
      }

      $form['create_payment']['payment_amount'] = [
        '#type' => 'number',
        '#title' => t('Amount') . ' ' . $currency_symbol,
        '#size' => 10,
        '#default_value' => $this->default_amount,
        '#min' => 1,
        '#step' => .01,
      ];

      $form['create_payment']['payment_title'] = [
        '#type' => 'textfield',
        '#title' => t('Payment title'),
        '#default_value' => t('Payment for @default_name', [
          '@default_name' => $this->default_name,
        ]),
      ];

      $form['create_payment']['charge_date'] = [
        '#title' => t('Charge customer at'),
        '#type' => 'date',
        '#default_value' => date('Y-m-d', strtotime($this->mandate->next_possible_charge_date)),
      ];

      $form['create_payment']['payment_submit'] = [
        '#type' => 'submit',
        '#value' => 'Instruct GoCardless to create a Payment',
        '#validate' => [[$this, 'paymentCreateValidate']],
        '#submit' => [[$this, 'paymentCreateSubmit']],
        '#attributes' => [
          'onclick' => 'if (!confirm("Are you sure you want to create a payment with GoCardless?")) {return false;}',
        ],
      ];
    }

    if ($this->gc->type == 'S') {

      if ($this->subscription->status != 'cancelled') {

        // Update subscription section.
        $form['update_subscription'] = [
          '#type' => 'details',
          '#title' => t('Update subscription'),
          '#open' => FALSE,
        ];

        $form['update_subscription']['subscription_name'] = [
          '#type' => 'textfield',
          '#title' => t('Subscription name'),
          '#default_value' => $this->default_name,
        ];

        $form['update_subscription']['subscription_amount'] = [
          '#type' => 'number',
          '#title' => t('Amount') . ' ' . $currency_symbol,
          '#size' => 10,
          '#default_value' => $this->default_amount,
          '#min' => 1,
          '#step' => .01,
        ];

        if (isset($data['shipment'])) {
          $form['update_subscription']['shipment'] = [
            '#type' => 'number',
            '#title' => t('Shipping proportion %'),
            '#description' => t("The proportion of the order's total shipping amount to allocate to this item."),
            '#size' => 5,
            '#default_value' => round($data['shipment']['proportion'] * 100, 2),
            '#step' => .01,
            '#min' => 0,
          ];
        }

        $form['update_subscription']['subscription_submit'] = [
          '#type' => 'submit',
          '#value' => 'Update subscription details at GoCardless',
          '#submit' => [[$this, 'subscriptionUpdateSubmit']],
        ];

        // Cancel subscription button.
        $form['cancel_subscription'] = [
          '#type' => 'submit',
          '#value' => 'Cancel subscription at GoCardless',
          '#submit' => [[$this, 'subscriptionCancelSubmit']],
          '#attributes' => [
            'onclick' => 'if (!confirm("Are you sure you want to cancel this subscription?")) {return false;}',
          ],
        ];
      }
    }

    if ($this->gc->type == 'P') {

      // Update payment creation section.
      if ($this->gc->gc_mandate_status != 'cancelled') {
        $form['update_payment'] = [
          '#type' => 'details',
          '#title' => t('Update scheduled payments creation'),
          '#open' => FALSE,
          '#access' => $this->gc->type == 'P' ? TRUE : FALSE,
        ];

        $form['update_payment']['next_payment'] = [
          '#type' => 'datetime',
          '#default_value' => $this->gc->next_payment ? DrupalDateTime::createFromTimestamp($this->gc->next_payment) : NULL,
          '#date_year_range' => '0:+1',
          // '#datepicker_options' => ['minDate' => 0],.
          '#description' => t('Change the date that the next scheduled payment will be created. (This is not the same date that the customer will be charged on.)'),
        ];

        $form['update_payment']['interval_length'] = [
          '#type' => 'number',
          '#title' => t('Interval length'),
          '#size' => 3,
          '#maxlength' => 3,
          '#default_value' => isset($data['interval_params']) ? $data['interval_params']['length'] : NULL,
          '#min' => 1,
        ];

        $form['update_payment']['interval_unit'] = [
          '#type' => 'select',
          '#title' => t('Interval unit'),
          '#default_value' => isset($data['interval_params']) ? $data['interval_params']['unit'] : NULL,
          '#options' => [
            'weekly' => t('Week'),
            'monthly' => t('Month'),
            'yearly' => t('Year'),
          ],
          '#empty_option' => t('- select -'),
        ];

        $form['update_payment']['amount'] = [
          '#type' => 'number',
          '#title' => t('Amount') . ' ' . $commerce_currency_symbol,
          '#size' => 5,
          '#default_value' => number_format((float) $total, 2, '.', ''),
          '#step' => .01,
          '#min' => 0,
          '#disabled' => TRUE,
        ];

        if (isset($data['shipment'])) {
          $form['update_payment']['shipment'] = [
            '#type' => 'number',
            '#title' => t('Shipping proportion %'),
            '#description' => t("The proportion of the order's total shipping amount to allocate to this item."),
            '#size' => 5,
            '#default_value' => round($data['shipment']['proportion'] * 100, 2),
            '#step' => .01,
            '#min' => 0,
          ];
        }

        $form['update_payment']['update_payment_button'] = [
          '#type' => 'submit',
          '#value' => 'Update payment',
          '#validate' => [[$this, 'paymentUpdateValidate']],
          '#submit' => [[$this, 'paymentUpdateSubmit']],
        ];
      }

      // Schedule adjustments section.
      $form['adjust'] = [
        '#type' => 'details',
        '#title' => t('Schedule adjustments'),
        '#description' => t("Adjust the amount of scheduled payment creations for item."),
        '#open' => isset($_GET['adjustment']) ? TRUE : FALSE,
      ];
      if ($this->gc->gc_mandate_status == 'cancelled' && !isset($adj_rows)) {
        $form['adjust']['#access'] = FALSE;
      }

      if (isset($adj_rows)) {
        $form['adjust']['adjust_table'] = [
          '#type' => 'details',
          '#title' => t('Adjustments'),
          '#open' => isset($_GET['adjustment']) ? TRUE : FALSE,
        ];
        $form['adjust']['adjust_table']['table'] = [
          '#type' => 'table',
          '#header' => $adj_headers,
          '#rows' => isset($adj_rows) ? $adj_rows : NULL,
        ];
      }

      if ($this->gc->gc_mandate_status != 'cancelled') {
        $form['adjust']['adjust_title'] = [
          '#type' => 'textfield',
          '#size' => 24,
          '#title' => t('Adjustment title'),
          '#default_value' => t('Adjustment for @default_name', [
            '@default_name' => $this->default_name,
          ]),
        ];

        $form['adjust']['adjustment'] = [
          '#title' => t('Adjustment amount') . ' ' . $currency_symbol,
          '#size' => 10,
          '#type' => 'number',
          '#step' => .01,
          '#default_value' => 0,
          '#description' => t('This will adjust the amount of a payment created with GoCardless.'),
        ];

        $form['adjust']['payments'] = [
          '#type' => 'number',
          '#size' => 6,
          '#title' => t('Number of payments'),
          '#default_value' => 1,
          '#min' => 1,
        ];

        $form['adjust']['starting_radio'] = [
          '#type' => 'radios',
          '#title' => t('Starting from'),
          '#options' => [
            0 => t('Next scheduled payment creation day'),
            1 => t('Select another date'),
          ],
          '#default_value' => 0,
        ];

        $form['adjust']['starting'] = [
          '#type' => 'date',
          '#title' => t('Starting from'),
          '#description' => t('The adjustment will begin on the first scheduled billing date after that specified here.'),
          // '#date_year_range' => '0:+1',
          // '#datepicker_options' => ['minDate' => 0],.
          '#states' => [
            'visible' => [
              ':input[name="starting_radio"]' => [
                'value' => 1,
              ],
            ],
          ],
        ];

        $form['adjust']['plus'] = [
          '#type' => 'details',
          '#title' => t('and then'),
          '#open' => FALSE,
          '#description' => t("Additional scheduled adjustment(s) to follow initial adjustment(s)."),
        ];

        $form['adjust']['plus']['plus_adjustment'] = [
          '#title' => t('Adjustment amount') . ' ' . $currency_symbol,
          '#size' => 10,
          '#type' => 'number',
          '#step' => .01,
        ];

        $form['adjust']['plus']['plus_payments'] = [
          '#type' => 'number',
          '#size' => 6,
          '#title' => t('Number of payments'),
          '#default_value' => NULL,
          '#min' => 1,
        ];

        $form['adjust']['adjust_button'] = [
          '#type' => 'submit',
          '#value' => 'Schedule',
          '#validate' => [[$this, 'adjustmentValidate']],
          '#submit' => [[$this, 'adjustmentSubmit']],
        ];
      }
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {}

  /**
   * Create Payment form validation.
   */
  public function paymentCreateValidate(array &$form, FormStateInterface $form_state) {

    // Check that specified date is greater than or equal to the next possible
    // charge date.
    if (!empty($form_state->getValue('charge_date'))) {
      if (strtotime($form_state->getValue('charge_date')) < strtotime($this->mandate->next_possible_charge_date)) {
        $form_state->setErrorByName('charge_date', t("The date cannot be before the 'Next Possible Charge Date'."));
      }
    }
  }

  /**
   * Create Payment form submission.
   */
  public function paymentCreateSubmit(array &$form, FormStateInterface $form_state) {

    $amount = $form_state->getValue('payment_amount');
    !empty($form_state->getValue('payment_title')) ? $title = $form_state->getValue('payment_title') : $title = $this->default_title;

    $result = $this->partner->api([
      'endpoint' => 'payments',
      'action' => 'create',
      'mandate' => $this->mandate->id,
      'amount' => $amount,
      'currency' => $this->currency_code,
      'description' => $title,
      'charge_date' => $form_state->getValue('charge_date'),
      'metadata' => ['item_id' => $this->item_id],
    ]);

    if (!isset($result->response)) {
      drupal_set_message(t('There was a problem creating the payment with GoCardless'), 'warning');
      return;
    }
    elseif ($result->response->status_code == 201) {
      $payment = $result->response->body->payments;
      drupal_set_message(t('Payment created successfully with GoCardless'));
    }

    // Dispatch an event so that other modules can respond to payment creation.
    $event = new PaymentCreatedEvent($payment, $this->item_id, $context = 'manual');
    $event_dispatcher = \Drupal::service('event_dispatcher');
    $event_dispatcher->dispatch(GoCardlessEvents::PAYMENT_CREATED, $event);
  }

  /**
   * Update Payment form validation.
   */
  public function paymentUpdateValidate(array &$form, FormStateInterface $form_state) {

    $interval_length = $form_state->getValue('interval_length');
    $interval_unit = $form_state->getValue('interval_unit');
    if (empty($interval_length) && !empty($interval_unit)) {
      $form_state->setErrorByName('interval_length', t("An interval length must be selected as well as the interval unit."));
    }
    if (!empty($interval_length) && empty($interval_unit)) {
      $form_state->setErrorByName('interval_unit', t("An interval unit must be selected as well as the interval length."));
    }
  }

  /**
   * Update Payment form submission.
   */
  public function paymentUpdateSubmit(array &$form, FormStateInterface $form_state) {

    $next_payment = strtotime($form_state->getValue('next_payment'));
    $this->db->update('commerce_gc_client_item')
      ->fields([
        'next_payment' => $next_payment ? $next_payment : NULL,
      ])
      ->condition('item_id', $this->item_id)
      ->execute();

    $interval_length = $form_state->getValue('interval_length');
    $interval_unit = $form_state->getValue('interval_unit');
    $gc = $this->item->getData('gc');
    if (!empty($interval_length)) {
      $gc['interval_params'] = [
        'length' => $interval_length,
        'unit' => $interval_unit,
        'string' => $interval_length . ' ' . str_replace("ly", "", $interval_unit),
        'gc' => $interval_length . ' ' . $interval_unit,
      ];
    }

    $shipment = $form_state->getValue('shipment');
    if ($shipment || $shipment == 0) {
      $gc['shipment']['proportion'] = (float) $shipment / 100;
    }
    $this->item->setData('gc', $gc);
    $this->item->save();

    drupal_set_message(t('Scheduled Payments Creation has been updated'));
  }

  /**
   * Scheduled Adjustment validation.
   */
  public function adjustmentValidate(array &$form, FormStateInterface $form_state) {

    $adj = $form_state->getValue('adjustment');
    $plus_adj = $form_state->getValue('plus_adjustment');

    if ($adj == '' || $adj == '0') {
      $form_state->setErrorByName('adjustment', t("You must provide an adjustment value."));
    }

    if ($plus_adj == '0') {
      $form_state->setErrorByName('plus_adjustment', t("Adjustment value cannot be set to zero."));
    }

    if (empty($form_state->getValue('payments'))) {
      $form_state->setValue('payments', 1);
    }

    if (!empty($plus_adj) && empty($form_state->getValue('plus_payments'))) {
      $form_state->setValue('plus_payments', 1);
    }
  }

  /**
   * Scheduled Adjustment submission.
   */
  public function adjustmentSubmit(array &$form, FormStateInterface $form_state) {

    $ints = $this->item->getData('gc')['interval_params'];
    $unit_price = $this->item->getUnitPrice()->getNumber();
    $unit_currency_code = $this->item->getUnitPrice()->getCurrencyCode();

    $query = $this->db->select('commerce_gc_client', 'g');
    $query->join('commerce_gc_client_item', 'i', 'g.gcid=i.gcid');
    $starting = $query->fields('i', ['next_payment'])->condition('item_id', $this->item_id)->execute()->fetchField();

    if ($form_state->getValue('starting_radio') == '1') {
      $select_date = strtotime($form_state->getValue('starting'));
      while ($starting < $select_date) {
        $string_ = '+' . $ints['string'];
        $starting = strtotime($string_, $starting);
      }
    }

    // Create array containing scheduled dates.
    $inserts = [];
    $payments = $form_state->getValue('payments');
    $amount = $form_state->getValue('adjustment');
    for ($i = 0; $i < $payments; $i++) {
      $string_ = '+' . ($i * $ints['length']) . ' ' . str_replace('ly', '', $ints['unit']);
      $inserts[] = [
        'timestamp' => strtotime($string_, $starting),
        'amount' => $amount,
      ];
      $ending = strtotime($string_, $starting);
    }

    if (!empty($form_state->getValue('plus_adjustment'))) {

      // Create array containing additional scheduled dates.
      $plus_amount = $form_state->getValue('plus_adjustment');
      $plus_starting = strtotime('+' . $ints['string'], $ending);
      $plus_payments = $form_state->getvalue('plus_payments');
      for ($i = 0; $i < $plus_payments; $i++) {
        $string_ = '+' . ($i * $ints['length']) . ' ' . str_replace('ly', '', $ints['unit']);
        $inserts[] = [
          'timestamp' => strtotime($string_, $plus_starting),
          'amount' => $plus_amount,
        ];
      }
    }

    // Add schedules data to database.
    foreach ($inserts as $insert) {

      // Check validity of new scheduled adjustments and disallow if it will
      // cause a payment creation of less than one and not zero.
      $insert_date = date('d M Y', $insert['timestamp']);

      // Calculate sum of scheduled adjs for date.
      $query = $this->db->select('commerce_gc_client_item', 'i');
      $query->join('commerce_gc_client_item_schedule', 's', 'i.item_id = s.item_id');
      $scheds = $query->fields('s')
        ->condition('s.type', 'adj')
        ->condition('s.status', 1)
        ->condition('s.date', $insert_date)
        ->condition('i.item_id', $this->item_id)
        ->execute()
        ->fetchAll();

      $sum = 0;
      foreach ($scheds as $sched) {
        $sched_data = unserialize($sched->data);
        $sum = $sum + $sched_data['amount'];
      }
      $sum = ($sum + $insert['amount'] + $unit_price);

      if ($sum < 1 && $sum != 0) {
        drupal_set_message(t('The schedule adjustment for @date cannot be placed because the price of the item inluding adjustments is not zero, and is less than @amount, which is not allowed by GoCardless.', [
          '@date' => $insert_date,
          '@amount' => \Drupal::service('commerce_price.currency_formatter')->format(1, $unit_currency_code),
        ]), 'warning');
        continue;
      }

      $insert_date = date('d M Y', $insert['timestamp']);
      $this->db->insert('commerce_gc_client_item_schedule')->fields([
        'item_id' => $this->item_id,
        'type' => 'adj',
        'date' => $insert_date,
        'timestamp' => strtotime($insert_date),
        'status' => 1,
        'data' => serialize([
          'title' => !empty($form_state->getValue('adjust_title')) ? $form_state->getValue('adjust_title') : t('Adjustment'),
          'amount' => $insert['amount'],
        ]),
        'created' => REQUEST_TIME,
      ])->execute();

      drupal_set_message(t('Schedule adjustment created for @date.', [
        '@date' => $insert_date,
      ]));
    }

    $form_state->setRedirect('commerce_gc_client.mandate', ['commerce_order' => $this->order_id], ['query' => ['adjustment' => 'created']]);
  }

  /**
   * Subscription Update submit.
   */
  public function subscriptionUpdateSubmit(array &$form, FormStateInterface $form_state) {

    $shipment = $form_state->getValue('shipment');
    if ($shipment || $shipment == 0) {
      $gc = $this->item->getData('gc');
      if ($gc['shipment']['proportion'] != (float) $shipment / 100) {
        $gc['shipment']['proportion'] = (float) $shipment / 100;
        $this->item->setData('gc', $gc);
        $this->item->save();
        drupal_set_message(t('Shipping proportion has been updated for the item.'));
      }
    }

    // Only update the subscription at GoCardless if the amount or the name
    // have been changed, since this action can only be done a limited number
    // of times.
    $amount = $form_state->getValue('subscription_amount');
    $name = $form_state->getValue('subscription_name');
    if ($amount * 100 != $this->subscription->amount || $name != $this->default_name) {
      $result = $this->partner->api([
        'endpoint' => 'subscriptions',
        'action' => 'update',
        'id' => $this->gc->gc_subscription_id,
        'amount' => $form_state->getValue('subscription_amount') * 100,
        'name' => $form_state->getValue('subscription_name'),
      ]
      );
      if ($result->response->status_code == 200) {
        drupal_set_message(t('Subscription was updated successfully at GoCardless.'));
      }
      else {
        drupal_set_message(t('There was a problem updating the subscription at GoCardless.'), 'warning');
      }
    }
  }

  /**
   * Subscription Canceltion submission.
   */
  public function subscriptionCancelSubmit(array &$form, FormStateInterface $form_state) {

    $result = $this->partner->api([
      'endpoint' => 'subscriptions',
      'action' => 'cancel',
      'id' => $this->gc->gc_subscription_id,
    ]
    );
    if ($result->response->status_code == 200) {
      drupal_set_message(t('Subscription was cancelled successfully. Please review any pending payments and cancel if necessary.'));
    }
    else {
      drupal_set_message(t('There was a problem cancelling the subscription with GoCardless.'), 'warning');
    }
  }

  /**
   * AJAX callback function.
   *
   * Stores the item ID as a session variable before returing form.
   */
  public static function switchItem(array &$form, FormStateInterface $form_state) {
    $item_id = $form_state->getValue('item_id');
    $order_id = $form_state->getValue('order_id');
    $session = \Drupal::service('user.private_tempstore')->get('commerce_gc_client');
    $session->set('item_id_' . $order_id, $item_id);
    $path = '/admin/commerce/orders/' . $order_id . '/gocardless';
    $response = new AjaxResponse();
    $response->addCommand(new RedirectCommand($path));
    return $response;
  }

}
