<?php

namespace Drupal\commerce_gc_client\Plugin\Commerce\PaymentGateway;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Ajax\RedirectCommand;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Database\Database;
use Symfony\Component\HttpFoundation\Request;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_gc_client\Controller\GoCardlessPartner;
use Drupal\commerce_gc_client\Event\GoCardlessEvents;
use Drupal\commerce_gc_client\Event\PaymentCreatedEvent;
use Drupal\commerce_gc_client\Event\PaymentNextEvent;
use Drupal\commerce_gc_client\Event\PaymentDetailsEvent;
use Drupal\commerce_gc_client\Event\SubsDetailsEvent;

/**
 * Provides the GoCardless Client payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "gocardless_client",
 *   label = "GoCardless Client",
 *   display_label = "Pay by Direct Debit with GoCardless",
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_gc_client\PluginForm\PaymentOffsiteForm",
 *   },
 *   modes = {
 *     "sandbox" = "Sandbox",
 *     "live" = "Live",
 *   },
 * )
 */
class GoCardlessClient extends OffsitePaymentGatewayBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'currencies' => FALSE,
      'fixer' => '',
      'payment_limit' => 3,
      'email_warnings' => \Drupal::config('system.site')->get('mail'),
      'log_webhook' => FALSE,
      'log_api' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildConfigurationForm($form, $form_state);
    $mode = $this->configuration['mode'];
    $module_settings = \Drupal::config('commerce_gc_client.settings');

    $connected = FALSE;
    if ($form_state->getValue('id') && $module_settings->get('org_id_' . $mode)) {
      $partner = new GoCardlessPartner($form_state->getValue('id'));
      $auth = $partner->api([
        'mode' => $mode,
      ]);
      if ($auth == 200) {
        $connected = TRUE;
      }
    }

    $connected ? $connect_submit = 'Disconnect' : $connect_submit = 'Connect';
    $mode == 'sandbox' ? $connect_value = $connect_submit . ' SANDBOX' : $connect_value = $connect_submit . ' LIVE';

    $form['mode']['#ajax'] = ['callback' => [$this, 'modeCallback']];

    $form['connect'] = [
      '#type' => 'fieldset',
      '#title' => t('Connect with GoCardless'),
    ];

    $markup = '<br /><p><b>Connect / disconnect with GoCardless</b></p>';
    if (!$connected) {
      $markup .= "<p>After clicking 'Connect' you will be redirected to the GoCardless where you can create an account and connect your site as a client of Seamless-CMS.co.uk</p>";
      $form['connect']['markup'] = [
        '#markup' => $markup,
      ];
    }

    if (!$form_state->getvalue('id')) {
      $connect_disabled = TRUE;
      $connect_suffix = t('<br /><i>Save the form before you can connect.</i>');
    }
    elseif ($mode == 'live' && !isset($_SERVER['HTTPS'])) {
      $connect_disabled = TRUE;
      $connect_suffix = t('<br /><i>Site needs to be secure (https) before you can connect to GoCardless LIVE.</i>');
    }

    $form['connect']['submit'] = [
      '#type' => 'submit',
      '#value' => $connect_value,
      '#submit' => [[$this, 'submit' . $connect_submit]],
      '#disabled' => isset($connect_disabled) ? $connect_disabled : FALSE,
      '#suffix' => isset($connect_suffix) ? $connect_suffix : NULL,
      '#attributes' => $connected ? ['onclick' => 'if (!confirm("Are you sure you want to disconnect your site from GoCardless?")) {return false;}'] : NULL,
    ];

    global $base_url;
    $webhook_url = $base_url . '/gc_client/webhook';

    $gc_webhook_url = t('https://manage@env.gocardless.com/developers/webhook-endpoints/create', [
      '@env' => $mode == 'sandbox' ? '-sandbox' : '',
    ]);

    $webhook_prefix = t('To receive webhooks, add <i>@webhook_url</i> as webhook URL and set the secret to the same as the Webhook Secret field from <a target="new" href="@gc_webhook_url">here</a>.<br /><br />', ['@webhook_url' => $webhook_url, '@gc_webhook_url' => $gc_webhook_url]);

    $form['connect']['webhook_secret'] = [
      '#type' => 'textfield',
      '#title' => 'Webhook secret',
      '#default_value' => $module_settings->get('webhook_secret_' . $mode),
      '#field_prefix' => $webhook_prefix,
    ];

    // Global.
    $form['global'] = [
      '#type' => 'fieldset',
      '#title' => t('General settings'),
    ];

    $form['global']['currencies'] = [
      '#type' => 'checkbox',
      '#title' => t('Allow customer to choose currency at GoCardless'),
      '#default_value' => $this->configuration['currencies'],
      '#description' => t('When directed to GoCardless during checkout, the customer is limited to the default Commerce currency, or the currency of the produce in their cart. If this option is enabled then the customer will be able to choose from all the currencies that you have enabled at GoCardless for their new direct debit mandate. See <a href = "/admin/commerce/config/currencies/gocardless">here</a> for more options, and information on enabling multiple currencies with GoCardless.'),
    ];

    $form['global']['fixer'] = [
      '#type' => 'textfield',
      '#title' => t('fixer.io API KEY'),
      '#default_value' => $this->configuration['fixer'],
      '#description' => "If you provide an API key, and the customer has chosen to set up their direct debit using a differnt currency, then during payment creation the amount will be automaticically calculated using current exchange rates at <a target='new' href='http://fixer.io'>fixer.io</a>. If you do not provide an API key payments will be the standard amount for the products the customer has purchased, regardless of the currency that the direct debit mandate was created to use.",
      '#size' => 40,
      '#maxlength' => 40,
      '#states' => [
        'disabled' => [
          'input[name="configuration[gc_client][global][currencies]"]' => ['checked' => FALSE],
        ],
      ],
    ];

    $form['global']['payment_limit'] = [
      '#type' => 'number',
      '#title' => t('Maximum payments'),
      '#default_value' => $this->configuration['payment_limit'],
      '#size' => 3,
      '#min' => 1,
      '#step' => 1,
      '#description' => t("The maximum number of payments that can be raised automatically, per order, per day. If the amount is exceeded, a warning email is sent to the specified address above. Leave unset for unlimitted."),
      '#required' => FALSE,
    ];

    $form['global']['email_warnings'] = [
      '#type' => 'textfield',
      '#title' => t('Email'),
      '#default_value' => $this->configuration['email_warnings'],
      '#description' => "Email address to send warnings.",
      '#size' => 40,
      '#maxlength' => 40,
    ];

    // Logging options.
    $form['log'] = [
      '#type' => 'fieldset',
      '#title' => t('Logging'),
    ];

    $form['log']['log_webhook'] = [
      '#prefix' => '<p>',
      '#suffix' => '</p>',
      '#type' => 'checkbox',
      '#title' => t('<b>Enable webhook logging</b>'),
      '#description' => 'Webhooks recieved from GoCardless will be written to the log.',
      '#default_value' => $this->configuration['log_webhook'],
    ];

    $form['log']['log_api'] = [
      '#prefix' => '<p>',
      '#suffix' => '</p>',
      '#type' => 'checkbox',
      '#title' => t('<b>Enable API logging</b>'),
      '#description' => 'Responses from the Partner site to API posts will be written to the log.',
      '#default_value' => $this->configuration['log_api'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {

    $mandate_id = $_GET['mandate'];
    $order_id = $order->id();
    $payment_gateway = $order->get('payment_gateway')->first()->entity;
    $payment_gateway_id = $payment_gateway->Id();
    $partner = new GoCardlessPartner($payment_gateway_id);

    $result = $partner->api([
      'endpoint' => 'mandates',
      'action' => 'get',
      'mandate' => $mandate_id,
    ]);

    if ($result->response->status_code != 200) {
      throw new PaymentGatewayException('GoCardless mandate has not been created for order.');
    }
    else {
      $mandate = $result->response->body->mandates;
    }

    // Insert info about the order into the database.
    $mode = $payment_gateway->get('configuration')['mode'];
    $db = \Drupal::database();
    $gcid = $db->insert('commerce_gc_client')->fields([
      'order_id' => $order_id,
      'gc_mandate_id' => $mandate_id,
      'gc_mandate_scheme' => $mandate->scheme,
      'gc_mandate_status' => 'pending_submission',
      'created' => REQUEST_TIME,
      'data' => serialize([
            // 'gc_redirect_id' => $_GET['redirect'],
    // TODO.
        'gc_redirect_id' => NULL,
        'gc_customer_id' => $_GET['customer'],
      ]),
        // 'data' => serialize($gc_data),
      'sandbox' => $mode == 'sandbox' ? 1 : 0,
    ])->execute();

    $comment = t('Your new direct debit mandate @mandate has been created by GoCardless.', ['@mandate' => $mandate_id]);
    drupal_set_message($comment);

    $event_dispatcher = \Drupal::service('event_dispatcher');

    // Set up seperate payments / subscriptions for each product item in order.
    foreach ($order->getItems() as $item) {

      $item_id = $item->Id();
      if (!$data = $item->getData('gc')) {
        $data = [];
      }
      if (empty($data)) {
        $data['gc_type'] = 'P';
      }

      // Payment type is subscription.
      if ($data['gc_type'] == 'S') {

        $calculate = commerce_gc_client_price_calculate($order, $item);

        !empty($data['gc_start_date']) ? $start_date = \Drupal::service('date.formatter')->format(strtotime($data['gc_start_date']), 'gocardless') : $start_date = NULL;
        !empty($data['gc_end_date']) ? $end_date = \Drupal::service('date.formatter')->format(strtotime($data['gc_end_date']), 'gocardless') : $end_date = NULL;

        $subs_details = [
          'endpoint' => 'subscriptions',
          'action' => 'create',
          'mandate' => $mandate_id,
          'amount' => $calculate['amount'],
          'currency' => $calculate['currency_code'],
          'name' => t('Subscription for') . ' ' . $item->getTitle(),
          'interval' => $data['interval_params']['length'],
          'interval_unit' => $data['interval_params']['unit'],
          'start_date' => $start_date,
          'count' => !empty($data['gc_count']) ? $data['gc_count'] : NULL,
          'end_date' => $end_date,
          'day_of_month' => !empty($data['gc_dom']) ? $data['gc_dom'] : NULL,
          'month' => !empty($data['gc_month']) ? $data['gc_month'] : NULL,
          'metadata' => [
            'item_id' => $item_id,
          ],
        ];

        // Dispatch an event so that the subscription details array can be
        // altered by other modules before sending to GoCardless.
        $subs_details_event = new SubsDetailsEvent($subs_details, $item_id);
        $event_dispatcher->dispatch(GoCardlessEvents::SUBS_DETAILS, $subs_details_event);
        $subs_details = $subs_details_event->getSubsDetails();

        $result = $partner->api($subs_details);

        if ($result->response->status_code == 201) {

          $sub = $result->response->body->subscriptions;
          $amount = \Drupal::service('commerce_price.currency_formatter')
            ->format($sub->amount / 100, $calculate['currency_code']);

          $comment_arr = [
            '@product' => $item->getTitle(),
            '@interval' => $sub->interval,
            '@interval_unit' => $sub->interval_unit,
            '@amount' => $amount,
            '@first_payment' => format_date(strtotime(array_shift($sub->upcoming_payments)->charge_date), 'gocardless_client'),
          ];

          $comment = t('Your @interval @interval_unit subscription of @amount for <b>@product</b> has been created with GoCardless, and the first payment will be made from your bank on @first_payment.', $comment_arr);
          drupal_set_message($comment);

          $db->insert('commerce_gc_client_item')->fields([
            'gcid' => $gcid,
            'item_id' => $item_id,
            'type' => $data['gc_type'],
            'gc_subscription_id' => $sub->id,
          ])->execute();
        }
        else {
          drupal_set_message(t('Something went wrong creating your subscription with GoCardless. Please contact the site administrator for assistance.'), 'warning');
        }
      }

      // Payment type is payment.
      elseif ($data['gc_type'] == 'P') {

        if (isset($next_payment)) {
          unset($next_payment);
        }

        $db->insert('commerce_gc_client_item')->fields([
          'item_id' => $item_id,
          'gcid' => $gcid,
          'type' => $data['gc_type'],
        ])->execute();

        if (count($data) == 1) {
          $first_payment = 1;
        }
        else {
          $first_payment = $this->startDateCalculate($data);
        }

        // Dispatch an event so that the first_payment date can be altered by
        // other modules.
        $first_payment_event = new PaymentNextEvent($first_payment, $item_id, $context = 'first_payment');
        $event_dispatcher->dispatch(GoCardlessEvents::PAYMENT_NEXT, $first_payment_event);
        $first_payment = $first_payment_event->getNextPayment();

        // If the product is set to create the first payment immediately.
        if ($first_payment == 1) {

          $calculate = commerce_gc_client_price_calculate($order, $item);

          $payment_details = [
            'endpoint' => 'payments',
            'action' => 'create',
            'mandate' => $mandate_id,
            'amount' => $calculate['amount'],
            'currency' => $calculate['currency_code'],
            'description' => t('Payment for') . ' ' . $item->getTitle(),
            'metadata' => [
              'item_id' => $item_id,
            ],
          ];

          // Dispatch an event so that the payment details array can be altered
          // by other modules before sending to GoCardless.
          $payment_details_event = new PaymentDetailsEvent($payment_details, $item_id, $context = 'checkout');
          $event_dispatcher->dispatch(GoCardlessEvents::PAYMENT_DETAILS, $payment_details_event);
          $payment_details = $payment_details_event->getPaymentDetails();

          $result = $partner->api($payment_details);

          if ($result->response->status_code == 201) {

            $payment = $result->response->body->payments;
            if (isset($data['interval_params'])) {
              $next_payment = strtotime('+' . $data['interval_params']['string']);
            }

            $amount = \Drupal::service('commerce_price.currency_formatter')
              ->format($payment->amount / 100, $calculate['currency_code']);

            $comment_array = [
              '@product' => $item->getTitle(),
              '@amount' => $amount,
              '@charge_date' => format_date(strtotime($result->response->body->payments->charge_date), 'gocardless_client'),
            ];
            $comment = t('A payment of @amount for <b>@product</b> has been created and will be made from your bank on @charge_date.', $comment_array);
            drupal_set_message($comment);

            // Dispatch an event so that other modules can respond to payment
            // creation.
            $payment_created_event = new PaymentCreatedEvent($payment, $item_id, $op = 'checkout');
            $event_dispatcher->dispatch(GoCardlessEvents::PAYMENT_CREATED, $payment_created_event);
          }
          else {
            drupal_set_message(t('There was a problem creating your initial payment. We will try to create it again later.'), 'warning');

            // Update next_payment to current time so that it will get picked
            // up on next cron run.
            $next_payment = REQUEST_TIME;
          }
        }

        // Else if a first_payment date is set for the product then defer the
        // first payment creation.
        elseif ($first_payment) {
          $next_payment = $first_payment;
          $calculate = commerce_gc_client_price_calculate($order, $item);

          $initial_payment = \Drupal::service('commerce_price.currency_formatter')->format($calculate['amount'] / 100, $calculate['currency_code']);
          $comment = t('A payment for @initial_payment will be created with GoCardless on @charge_date.', [
            '@initial_payment' => $initial_payment,
            '@charge_date' => format_date(strtotime('midnight', $first_payment), 'gocardless_client'),
          ]);
          drupal_set_message($comment);
        }

        // Update next_payment value in database.
        if (isset($next_payment)) {

          // Dispatch an event so that the next_payment date can be altered by
          // other modules.
          $next_payment_event = new PaymentNextEvent($next_payment, $item_id, $context = 'checkout');
          $event_dispatcher->dispatch(GoCardlessEvents::PAYMENT_NEXT, $next_payment_event);
          $next_payment = $next_payment_event->getNextPayment();

          $db->update('commerce_gc_client_item')->fields([
            'next_payment' => $next_payment,
          ])->condition('item_id', $item_id)->execute();
        }
      }
      $item->save();
    }
    $order->save();
  }

  /**
   * Calculates first payment creation date for an order item.
   *
   * Returns TRUE if the product variation is set to create a payment
   * immediately, or a Unix timestamp if a first payment date has been
   * calculated.
   *
   * @param array $data
   *   The data array for the new order item.
   * @param string $date
   *   (optional) A date as a string. For testing purposes only.
   *
   * @return mixed
   *   TRUE or integer.
   */
  public static function startDateCalculate(array $data, $date = FALSE) {

    if ($data['gc_create_payment']) {
      return TRUE;
    }

    $date ? $time = strtotime($date) : $time = strtotime('midnight');
    $data['gc_start_date'] ? $start_time = strtotime($data['gc_start_date']) : $start_time = $time;

    // Weekly intervals.
    if ($data['gc_interval_unit'] == 'weekly') {
      $dow = date('l', $start_time);
      if ($dow == $data['gc_dow'] || !$data['gc_dow']) {
        $start_time > $time ? $fp = $start_time : $fp = $time;
      }
      else {
        $start_time < $time ? $start_time = $time : NULL;
        $fp = strtotime('next ' . $data['gc_dow'], $start_time);
      }
      return ($fp);
    }

    // Monthly intervals.
    elseif ($data['gc_interval_unit'] == 'monthly') {
      $data['gc_dom'] == -1 ? $string = 'last day of ' : $string = $data['gc_dom'];
      $start_time < $time ? $start_time = $time : NULL;
      $m = date('M', $start_time);
      $y = date('Y', $start_time);
      $fp = strtotime($string . $m . $y);
      if ($fp < $start_time) {
        $fp = strtotime('+1 month', $fp);
      }
      return $fp;
    }

    // Yearly intervals.
    elseif ($data['gc_interval_unit'] == 'yearly') {
      $data['gc_dom'] == -1 ? $string = 'last day of ' : $string = $data['gc_dom'];
      $y = date('Y', $start_time);
      $fp = strtotime($string . $data['gc_month'] . $y);
      if ($fp < $start_time) {
        $fp = strtotime('+1 year', $fp);
        // Take account of leap years.
        if (date('n', $fp) == 2 && $data['gc_dom'] == -1) {
          $fp = strtotime('last day of ' . date('M', $fp), $fp);
        }
      }
      return $fp;
    }

    // No interval set.
    elseif ($data['gc_start_date']) {
      $fp = strtotime($data['gc_start_date']);
      if ($fp > $time) {
        return $fp;
      }
    }

    return $time;
  }

  /**
   * Redirect to Partner site to activate the GoCardless OAuth Flow.
   */
  public function submitConnect(array &$form, FormStateInterface $form_state) {

    global $base_url;
    $gateway_id = $form_state->getvalue('id');
    $client_url = urlencode($base_url . '/gc_client/connect_complete');
    $site_email = \drupal::config('system.site')->get('mail');
    $env = $form_state->getvalue('configuration')['mode'];
    $partner_url = \Drupal::config('commerce_gc_client.settings')->get('partner_url');
    $token = user_password($length = 24);
    \Drupal::service('config.factory')->getEditable('commerce_gc_client.settings')->set('partner_pass_' . $env, $token)->save();
    $url = $partner_url . '/gc_partner/connect/' . $env . '?mail=' . $site_email . '&client_url=' . $client_url . '&gateway_id=' . $gateway_id . '&token=' . $token . '&module=Commerce 8.x';

    $response = new trustedredirectresponse($url);
    $form_state->setresponse($response);

    // Remove the 'desination=' parameter from query string if it exists.
    $request = \Drupal::request();
    $request->query->set('destination', NULL);
  }

  /**
   * Disconnects client site from GC partner site.
   */
  public function submitDisconnect(array &$form, FormStateInterface $form_state) {

    $id = $form_state->getValue('id');
    $partner = new GoCardlessPartner($id);
    $result = $partner->api([
      'endpoint' => 'oauth',
      'action' => 'revoke',
    ]);

    if ($result->response == 200) {
      drupal_set_message(t('You have disconnected successfully from GoCardless'));
    }
    else {
      drupal_set_message(t('There was a problem disconnecting from GoCardless'), 'error');
    }

    // TODO The following should only happen if there are no other GoC payment
    // gateways installed.
    $session = \Drupal::request()->getSession();
    if ($session->get('gc_client_cookie_created')) {
      $session->remove('gc_client_cookie_created');
    }
    $mode = $form_state->getValue('configuration')['mode'];
    $config = \Drupal::service('config.factory')->getEditable('commerce_gc_client.settings');
    $config->set('org_id_' . $mode, NULL)->save();
    $config->set('partner_user_' . $mode, NULL)->save();
    $config->set('partner_pass_' . $mode, NULL)->save();

    // Remove the 'desination=' parameter from query string if it exists.
    $request = \Drupal::request();
    $request->query->set('destination', NULL);
  }

  /**
   * Ajax callback function.
   *
   * On changing 'mode', saves new configuration and reloads the page.
   */
  public function modeCallback(array &$form, FormStateInterface $form_state) {

    $gateway_id = $form_state->getValue('id');
    $mode = $form_state->getValue('configuration')['gc_client']['mode'];

    \Drupal::configFactory()->getEditable('commerce_payment.commerce_payment_gateway.' . $gateway_id)->set('configuration.mode', $mode)->save();

    if (empty($gateway_id)) {
      $path = '/admin/commerce/config/payment-gateways/add';
    }
    else {
      $path = '/admin/commerce/config/payment-gateways/manage/' . $gateway_id;
    }

    $response = new AjaxResponse();
    $response->addCommand(new RedirectCommand($path));
    return $response;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {

    parent::submitConfigurationForm($form, $form_state);
    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['webhook_secret'] = $values['connect']['webhook_secret'];
      $this->configuration['currencies'] = $values['global']['currencies'];
      $this->configuration['fixer'] = $values['global']['fixer'];
      $this->configuration['email_warnings'] = $values['global']['email_warnings'];
      $this->configuration['payment_limit'] = $values['global']['payment_limit'];
      $this->configuration['log_api'] = $values['log']['log_api'];
      $this->configuration['log_webhook'] = $values['log']['log_webhook'];

      $mode = $this->configuration['mode'];
      $settings = \Drupal::service('config.factory')->getEditable('commerce_gc_client.settings');
      $settings->set('webhook_secret_' . $mode, $values['connect']['webhook_secret'])->save();
    }
  }

}
