<?php

namespace Drupal\commerce_gc_client\PluginForm;

use Drupal\commerce_shipping\Entity\Shipment;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Utility\UrlHelper;
use Drupal\commerce_gc_client\Controller\GoCardlessPartner;
use Drupal\commerce_gc_client\Event\GoCardlessEvents;
use Drupal\commerce_gc_client\Event\MandateDetailsEvent;
use Drupal\commerce_gc_client\Event\ShipmentEvent;

/**
 * Defines the payment gateway's off-site payment form.
 */
class PaymentOffsiteForm extends BasePaymentOffsiteForm {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {

    $event_dispatcher = \Drupal::service('event_dispatcher');
    $form = parent::buildConfigurationForm($form, $form_state);
    $payment = $this->entity;
    $payment_gateway_plugin = $payment->getPaymentGateway()->getPlugin();
    $mode = $payment_gateway_plugin->getMode();
    $order = $payment->getOrder();

    // Obtain shipping data for order.
    $moduleHandler = \Drupal::service('module_handler');
    if ($moduleHandler->moduleExists('commerce_shipping')) {
      $db = \Drupal::database();
      $shipment_id = $db->select('commerce_shipment', 's')
        ->fields('s', ['shipment_id'])
        ->condition('order_id', $order->id())
        ->execute()->fetchField();

      if ($shipment_id) {
        $shipment = Shipment::Load($shipment_id);
        foreach ($order->getItems() as $item) {

          // Dispatch an event so that the shipping price per item can be
          // added to the order item's data array.
          $event = new ShipmentEvent($shipment, $item);
          $event_dispatcher->dispatch(GoCardlessEvents::SHIPMENT, $event);

          if ($shipment_proportion = $event->getItemShipmentProportion()) {
            if (!$data = $item->getData('gc')) {
              $data = [];
            }
            $data['shipment'] = [
              'plugin_id' => $shipment->getShippingMethod()->getPlugin()->getPluginId(),
              'proportion' => $shipment_proportion,
            ];
            $item->setData('gc', $data)->save();
          }
        }
      }
    }

    $order_id = $order->Id();
    $partner = new GoCardlessPartner($this->entity->getPaymentGatewayId());
    $redirect_url = $partner->settings['partner_url'] . '/gc_partner/mandate?payment_method=gocardless_client&org_id=' . $partner->settings['org_id_' . $mode] . '&order_id=' . $order_id . '&complete_path=' . $form['#return_url'];

    // Generate the mandate_details array.
    $user = ['email' => $order->getEmail()];
    if ($profile = $order->getBillingProfile()) {
      $user += [
        'first_name' => $profile->address->given_name,
        'last_name' => $profile->address->family_name,
        'company_name' => $profile->address->organisation,
        'billing_address1' => $profile->address->address_line1,
        'billing_address2' => $profile->address->address_line2,
        'billing_town' => $profile->address->locality,
        'billing_postcode' => $profile->address->postal_code,
        'billing_country' => $profile->address->country_code,
      ];
      $user = array_map(function ($str) {
        return $str ? $str : '';
      }, $user);
    }

    $scheme = NULL;
    if (!$partner->settings['configuration']['currencies']) {
      $currency_code = $order->getTotalPrice()->getCurrencyCode();
      foreach ($partner->settings['currency_schemes'] as $currency_scheme => $value) {
        if ($currency_code == $currency_scheme) {
          $scheme = $value['scheme'];
          break;
        }
      }
    }

    $mandate_details = [
      'name' => 'Order ' . $order_id,
      'scheme' => $scheme,
      'redirect_url' => $redirect_url,
      'user' => $user,
    ];

    // Dispatch an event so that the mandate details array can be altered
    // by other modules before sending to GoCardless.
    $mandate_details_event = new MandateDetailsEvent($mandate_details, $order_id);
    $event_dispatcher->dispatch(GoCardlessEvents::MANDATE_DETAILS, $mandate_details_event);
    $mandate_details = $mandate_details_event->getMandateDetails();

    $result = $partner->api([
      'endpoint' => 'redirect_flows',
      'action' => 'create',
      'mandate_details' => $mandate_details,
      'order_id' => $order_id,
    ]);

    if (isset($result->response) && UrlHelper::isValid($result->response)) {
      $redirect_url = $result->response;
      $data = [];
      $redirect_method = 'redirect';
      $form = $this->buildRedirectForm($form, $form_state, $redirect_url, $data, $redirect_method);
    }
    else {
      drupal_set_message(t('Something went wrong during connection with GoCardless'), 'error');
    }

    return $form;
  }

}
